#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <limits.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <archive.h>
#include <archive_entry.h>

#include "arg.h"
#include "config.h"

#define log(l,...) if(verbose>=l){printf(__VA_ARGS__);}

struct pack {
	char *path;
	char *name;
	char *version;
};

/* possible actions */
enum {
	ACTION_INVALID     = -1,
	ACTION_INSTALL     = 0,
	ACTION_DELETE      = 1,
	ACTION_UPDATE      = 2,
	ACTION_INSPECT     = 3,
	ACTION_DEFAULT     = ACTION_INSPECT,
};

/* error codes */
enum {
	ERR_INVALID_ACTION = 1,
	ERR_METADATA       = 2,
	ERR_PACK_LOAD      = 3,
	ERR_DELETE         = 4,
	ERR_INSPECT        = 5,
	ERR_UNPACK         = 6,
};

enum {
	LOG_NONE    = 0,
	LOG_INFO    = 1,
	LOG_VERBOSE = 2,
	LOG_DEBUG   = 3
};

void usage(char *name);
int is_empty(char *dir);
int mkdir_parents(char *dir, mode_t mode);
char *base_name(char *path);
int re_match(const char *re, const char *str);

struct pack *pack_load_tarball(char *path);
struct pack *pack_load_metadata(const char *datadir, char *name);
void pack_free(struct pack *p);
#ifdef REPOAWARE
int  pack_find(char *, char *);
#endif
int pack_extract(const char *rootfs, const char *datadir, struct pack *p);
int pack_install(const char *rootfs, const char *datadir, struct pack *p);
int pack_delete(const char *rootfs, const char *datadir, struct pack *p);

int inspect_installed(const char *datadir, const char *name);
int inspect_version(const char *datadir, const char *name, char version[]);
int inspect_collision(const char *rootfs, struct pack *p);
int inspect_files(int fd, const char *datadir, const char *name);
int inspect_system(int fd, const char *datadir);

int write_metadata(const char *datadir, struct pack *pack);
int write_entry(struct archive *a, struct archive *w);
int delete_node(char *path);
int delete_content(const char *rootfs, char *map, size_t size);
int delete_metadata(const char *datadir, char *name);

/* action wrappers around CLI arguments */
int install(const char *rootfs, const char *datadir, char *path);
int update(const char *rootfs, const char *datadir, char *path);
int delete(const char *rootfs, const char *datadir, char *name);
int inspect(const char *datadir, const char *name);

int verbose = LOG_NONE;
int overwrite = 0;

void
usage(char *name)
{
	fprintf(stderr, "usage: %s -adfiuv [PACK..]\n" , name);
}


/*
 * Returns 0 if a directory is empty, -1 otherwise
 */
int
is_empty(char *path)
{
	DIR *d;
	struct dirent *p;

	if (!(d = opendir(path))) {
		perror(path);
		return -1;
	}

	while ((p = readdir(d))) {
		if (strcmp(p->d_name, ".") && strcmp(p->d_name, "..")) {
			closedir(d);
			return -1;
		}
	}

	closedir(d);
	return 0;
}


/*
 * Recursive mkdir, taken from the ii project
 * http://nion.modprobe.de/blog/archives/357-Recursive-directory-creation.html
 */
int
mkdir_parents(char *path, mode_t mode)
{
        char tmp[PATH_MAX] = "";
        char *p = NULL;
        size_t len;

	snprintf(tmp, sizeof(tmp), "%s", path);
	len = strlen(tmp);
	if(tmp[len - 1] == '/')
		tmp[len - 1] = 0;
	for(p = tmp + 1; *p; p++)
		if(*p == '/') {
			*p = 0;
			mkdir(tmp, mode);
			*p = '/';
		}
	return mkdir(tmp, mode);
}


/*
 * Return a pointer to the basename, or NULL if path ends with '/'
 */
char *
base_name(char *path)
{
	char *b = strrchr(path, '/');
	return b ? b + 1 : path;
}


/*
 * Check wether a string matches a regular expression
 */
int
re_match(const char *re, const char *str)
{
	int retval = -1;
	regex_t preg;
	regmatch_t sub[1];

	if (regcomp(&preg, re, 0)) {
		fprintf(stderr, "%s: Invalid regular expression\n", re);
		return -1;
	}

	retval = regexec(&preg, str, 1, sub, 0);
	regfree(&preg);

	return retval;
}


/*
 * Load a pack from a tarball and return a pack structure
 */
struct pack *
pack_load_tarball(char *path)
{
	struct stat st;
	struct pack *pack = NULL;
	char *fn, *regex = PACK_FORMAT;
	regex_t preg;
	regmatch_t sub[3];
	size_t i, nmatch = 3, sublen[3];

	fn = base_name(path);
	if (stat(path, &st) < 0) {
		perror(path);
		return NULL;
	}

	regcomp(&preg, regex, 0);
	regexec(&preg, fn, nmatch, sub, 0);

	if (!(pack = malloc(sizeof(struct pack)))) {
		perror(path);
		regfree(&preg);
		return NULL;
	}

	for (i=0; i<nmatch; i++)
		sublen[i] = sub[i].rm_eo - sub[i].rm_so;

	pack->path    = strdup(path);
	pack->name    = malloc(sublen[1] + 1);
	pack->version = malloc(sublen[2] + 1);

	strncpy(pack->name, fn+sub[1].rm_so, sublen[1]);
	strncpy(pack->version, fn+sub[2].rm_so, sublen[2]);

	pack->name[sublen[1]] = 0;
	pack->version[sublen[2]] = 0;

	regfree(&preg);

	return pack;
}


/*
 * Load a pack from a metadata directory and return a pack structure
 */
struct pack *
pack_load_metadata(const char *datadir, char *name)
{
	struct pack *pack = NULL;
	char tmp[PATH_MAX] = "";

	snprintf(tmp, PATH_MAX, "%s/%s", datadir, name);

	if (inspect_installed(datadir, name)) {
		fprintf(stderr, "%s: No such pack installed\n", name);
		return NULL;
	}

	if (!(pack = malloc(sizeof(struct pack)))) {
		perror("malloc");
		return NULL;
	}

	pack->name = strdup(name);
	pack->path = strdup(tmp);
	pack->version = malloc(LINE_MAX);
	inspect_version(datadir, pack->name, pack->version);

	return pack;
}


/*
 * Free a pack structure
 */
void
pack_free(struct pack *p)
{
	if (!p)
		return;

	if (p->path)
		free(p->path);
	if (p->name)
		free(p->name);
	if (p->version)
		free(p->version);

	free(p);
}


#ifdef REPOAWARE
/*
 * Find a pack filename using an external tool writing the path to the
 * pack to stdout.
 * The tool should be called as:
 *
 *	tool <name>
 */
int
pack_find(char *name, char *out)
{
	int fd[2], status;
	size_t len = 0;

	pipe(fd);
	if (!fork()) {
		close(1);
		close(fd[0]);
		dup2(fd[1], 1);
		execlp(REPO_EXEC, REPO_EXEC, name, NULL);
	}

	close(fd[1]);

	wait(&status);
	if (status)
		exit(1);

	len = read(fd[0], out, PATH_MAX);
	close(fd[0]);

	if (len < 1)
		return -1;

	out[len - 1] = 0;

	return 0;
}
#endif


/*
 * Extract a tarball to the given directory
 */
int
pack_extract(const char *rootfs, const char *datadir, struct pack *p)
{
	int r, fd;
	struct archive *a;
	struct archive *w;
	struct archive_entry *e;
	char cwd[PATH_MAX] = "";
	char tmp[PATH_MAX] = "";

	int mask = ARCHIVE_EXTRACT_PERM
	          |ARCHIVE_EXTRACT_SECURE_NODOTDOT;

	snprintf(tmp, PATH_MAX, "%s/%s/files", datadir, p->name);

	log(LOG_DEBUG, "+ %s\n", tmp);
	if ((fd = open(tmp, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0) {
		perror(tmp);
		return -1;
	}

	a = archive_read_new();
	archive_read_support_filter_gzip(a);
	archive_read_support_filter_bzip2(a);
	archive_read_support_filter_xz(a);
	archive_read_support_format_tar(a);

	/* try and open the tarball of our pack */
	if ((r = archive_read_open_filename(a, p->path, 0)) != ARCHIVE_OK) {
		fprintf(stderr, "%s: %s\n", p->path, archive_error_string(a));
		archive_read_free(a);
		close(fd);
		return r;
	}

	w = archive_write_disk_new();
	archive_write_disk_set_options(w, mask);

	getcwd(cwd, PATH_MAX);
	chdir(rootfs);
	log(LOG_VERBOSE, "extracting pack\n");
	while ((r = archive_read_next_header(a, &e)) != ARCHIVE_EOF) {
		if (r != ARCHIVE_OK) {
			fprintf(stderr, "%s: %s\n", archive_entry_pathname(e),
			                            archive_error_string(a));
			break;
		}

		if ((r = archive_write_header(w, e)) != ARCHIVE_OK) {
			fprintf(stderr, "%s: %s\n", archive_entry_pathname(e),
			                            archive_error_string(w));
			break;
		}

		log(LOG_DEBUG, "+ %s%s\n", rootfs, archive_entry_pathname(e));
		if ((r = write_entry(a, w)) != ARCHIVE_OK) {
			fprintf(stderr, "%s: %s\n", archive_entry_pathname(e),
			                            archive_error_string(w));
			break;
		}

		dprintf(fd, "%s\n", archive_entry_pathname(e));
		archive_write_finish_entry(w);
	}
	close(fd);
	chdir(cwd);

	archive_write_free(w);
	archive_read_free(a);

	return r == ARCHIVE_EOF ? ARCHIVE_OK : r;
}


/*
 * Install a pack to rootfs, writing metadata to datadir
 * If overwrite is set to 1, it will overwrite all files
 */
int
pack_install(const char *rootfs, const char *datadir, struct pack *p)
{
	int r;
	char tmp[PATH_MAX] = "";

	if (overwrite == 0) {
		snprintf(tmp, PATH_MAX, "%s/%s", datadir, p->name);
		log(LOG_VERBOSE, "checking installation status\n");
		if (inspect_installed(datadir, p->name) == 0) {
			fprintf(stderr, "%s: Pack already installed\n", p->name);
			return -1;
		}
		log(LOG_VERBOSE, "checking for collisions\n");
		if (inspect_collision(rootfs, p) != 0)
			return -1;
	}

	log(LOG_VERBOSE, "writing metadata\n");
	if (write_metadata(datadir, p) < 0) {
		delete_metadata(datadir, p->name);
		return -1;
	}

	r = pack_extract(rootfs, datadir, p);

	if (r != 0) {
		fprintf(stderr, "%s: Installation failed. Deleting pack\n", p->name);
		pack_delete(rootfs, datadir, p);
	}

	return r;
}


/*
 * Delete a pack from the system.
 * This will read the file datadir/name/files, change directory to rootfs
 * and call delete_content()
 */
int
pack_delete(const char *rootfs, const char *datadir, struct pack *p)
{
	int fd, r = 0;
	char *addr = NULL;
	char tmp[PATH_MAX];
	struct stat st;

	snprintf(tmp, PATH_MAX, "%s/%s/files", datadir, p->name);
	if (stat(tmp, &st) < 0) {
		perror(tmp);
		return -1;
	}
	if ((fd = open(tmp, O_RDONLY)) < 0) {
		perror(tmp);
		return ERR_DELETE;
	}

	if (st.st_size > 0) {
		addr = mmap(0, st.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
		if (addr == MAP_FAILED) {
			perror("mmap");
			close(fd);
			return ERR_DELETE;
		}
	
		log(LOG_VERBOSE, "deleting installed files\n");
		/* ignore errors so everything gets deleted */
		if (delete_content(rootfs, addr, st.st_size) < 0) {
			fprintf(stderr, "%s: Pack removal failed\n", p->name);
			close(fd);
			r = ERR_DELETE;
		}
		munmap(addr, st.st_size);
	}
	close(fd);

	log(LOG_VERBOSE, "cleaning metadata\n");
	if (delete_metadata(datadir, p->name) < 0) {
		fprintf(stderr, "%s: Metadata removal failed\n", p->name);
		r = ERR_DELETE;
	}

	return r;
}


/*
 * Returns 0 if "name" is installed, non-zero otherwise
 */
int
inspect_installed(const char *datadir, const char *name)
{
	char tmp[PATH_MAX] = "";
	struct stat st;

	snprintf(tmp, PATH_MAX, "%s/%s", datadir, name);
	log(LOG_VERBOSE, "checking existence of %s\n", tmp);
	return stat(tmp, &st) || !S_ISDIR(st.st_mode);
}


/*
 * Read the version file for a pack and fill the given pointer with it
 */
int
inspect_version(const char *datadir, const char *name, char version[LINE_MAX])
{
	FILE *stream;
	char tmp[PATH_MAX] = "", *lf = NULL;

	snprintf(tmp, PATH_MAX, "%s/%s/version", datadir, name);
	if ((stream = fopen(tmp, "r")) == NULL) {
		sprintf(version, "(unknown)");
		return 1;
	} else {
		fgets(version, LINE_MAX, stream);
		if ((lf = strchr(version, '\n')) != NULL)
			*lf = 0;
	}

	fclose(stream);

	return 0;
}


/*
 * Check for collisions between the filesystem and the tarball
 */
int
inspect_collision(const char *rootfs, struct pack *p)
{
	int r = 0;
	char cwd[PATH_MAX] = "";
	struct stat st;
	struct archive *a;
	struct archive_entry *e;

	a = archive_read_new();
	archive_read_support_filter_gzip(a);
	archive_read_support_filter_bzip2(a);
	archive_read_support_filter_xz(a);
	archive_read_support_format_tar(a);

	r = archive_read_open_filename(a, p->path, 0);
	if (r != ARCHIVE_OK) {
		fprintf(stderr, "%s: %s\n", p->path, archive_error_string(a));
		return -1;
	}

	getcwd(cwd, PATH_MAX);
	chdir(rootfs);
	while (archive_read_next_header(a, &e) == ARCHIVE_OK) {
		if (stat(archive_entry_pathname(e), &st) == 0 && !S_ISDIR(st.st_mode)) {
			fprintf(stderr, "%s: File exists\n", archive_entry_pathname(e));
			r++;
			break;
		}
		archive_read_data_skip(a);
	}

	archive_read_free(a);
	chdir(cwd);

	return r;
}


/*
 * Write files installed by a pack to a file descriptor
 */
int
inspect_files(int fd, const char *datadir, const char *name)
{
	int meta;
	char tmp[PATH_MAX] = "";
	size_t len;

	snprintf(tmp, PATH_MAX, "%s/%s/files", datadir, name);
	if ((meta = open(tmp, O_RDONLY)) < 0) {
		perror(tmp);
		return -1;
	}

	while ((len = read(meta, tmp, PATH_MAX)) > 0) {
		tmp[len] = 0;
		dprintf(fd, "%s", tmp);
	}

	return 0;
}


/*
 * Write packs installed in datadir to a file descriptor
 */
int
inspect_system(int fd, const char *datadir)
{
	DIR *d;
	struct dirent *p;
	char ver[LINE_MAX];

	if (!(d = opendir(datadir))) {
		perror(datadir);
		return -1;
	}

	while ((p = readdir(d)))
		if (p->d_name[0] != '.') {
			inspect_version(datadir, p->d_name, ver);
			dprintf(fd, "%s\t%s\n", p->d_name, ver);
		}

	closedir(d);
	return 0;
}


/*
 * Write metadata about a pack file:
 * + datadir/packname/version - version of the pack installed
 */
int
write_metadata(const char *datadir, struct pack *p)
{
	int fd, r;
	struct stat st;
	char tmp[PATH_MAX];

	snprintf(tmp, PATH_MAX, "%s/%s", datadir, p->name);

	if (stat(tmp, &st) < 0 && errno == ENOENT) {
		log(LOG_DEBUG, "+ %s\n", tmp);
		if ((r = mkdir_parents(tmp, 0755)) < 0)
			return r;
	}

	snprintf(tmp, PATH_MAX, "%s/%s/version", datadir, p->name);

	log(LOG_DEBUG, "+ %s\n", tmp);
	if ((fd = open(tmp, O_CREAT|O_WRONLY|O_TRUNC, 0644)) < 0) {
		perror(tmp);
		return -1;
	}

	r  = write(fd, p->version, strnlen(p->version, LINE_MAX));
	r += write(fd, "\n", 1);

	if (r < 1) {
		perror(tmp);
		close(fd);
		return -1;
	}

	close(fd);
	return 0;
}


/*
 * Write an archive entry on the disk, thus creating the file
 */
int
write_entry(struct archive *a, struct archive *w)
{
	int r;
	const void *buf;
	size_t len;
	off_t off;

	for (;;) {
		r = archive_read_data_block(a, &buf, &len, &off);
		switch (r) {
		case ARCHIVE_EOF: return 0;
		case ARCHIVE_OK: break;
		default: return r;
		}

		r = archive_write_data_block(w, buf, len, off);
		if (r != ARCHIVE_OK)
			return r;
	}
}


/*
 * Deletes a node, be it a file or a directory.
 * In case the node doesn't exists, we're done
 */
int
delete_node(char *path)
{
	int r = 0;
	struct stat st;
	size_t len = 0;

	len = strnlen(path, PATH_MAX);
	/* remove potential trailing '/' */
	if (path[len - 1] == '/')
		path[--len] = 0;

	log(LOG_DEBUG, "- %s\n", path);

	/*
	 * if path doesn't exist anymore, it's all good :)
	 * we use lstat here so that dangling links can be delt with too
	 */
	if (lstat(path, &st) < 0 && errno == ENOENT)
		return 0;

	if (S_ISDIR(st.st_mode) && is_empty(path) == 0) {
		if ((r = rmdir(path)) < 0) {
			perror(path);
			return r;
		}
	}

	if (!S_ISDIR(st.st_mode) && (r = unlink(path)) < 0) {
		perror(path);
		return r;
	}

	return 0;
}


/*
 * Delete all entries listed in the given file.
 * if the entry doesn't exist, it will be ignored
 */
int
delete_content(const char *rootfs, char *map, size_t size)
{
	char *path = NULL;
	char tmp[PATH_MAX] = "";
	size_t off;

	if (size < 1)
		return -1;

	do {
		/*
		 * it is assumed here that the file is POSIX and thus that
		 * the last char will be \n.
		 * This might sound stupid; but this is what pack_extract() will do.
		 * We read the mmap file backward until we encounter \n or the
		 * beginning of the file. If it's an \n, we replace it by 0 and
		 * process to the deletion of the inode, either with rmdir or
		 * unlink.
		 *
		 * When setting the path, the current offset is either at the
		 * start of the file, or on a 0 (freshly replaced \n). In case
		 * We are on a 0, the path should be set to the string just
		 * after (&map[off] + 1). This doesn't apply to start of file.
		 */
		for (off=size-1; off>0 && map[off] != '\n'; off--);
		map[off] = (off > 0 ? 0 : *map);
		path     = (off < size-1 ? &map[off] + (off>0?1:0) : NULL);

		if (path != NULL && strnlen(path, PATH_MAX) > 0) {
			snprintf(tmp, PATH_MAX, "%s%s", rootfs, path);
			if (delete_node(tmp) < 0)
				return ERR_DELETE;
		}
	} while (off > 0);

	return 0;
}


/*
 * Delete the metadata stored for a given pack name
 */
int
delete_metadata(const char *datadir, char *name)
{
	int i;
	char path[PATH_MAX] = "";
	char *meta[] = { "files", "version", NULL };
	
	for (i = 0; meta[i] != NULL; i++) {
		snprintf(path, PATH_MAX, "%s/%s/%s", datadir, name, meta[i]);
		log(LOG_DEBUG, "- %s\n", path);
		if (unlink(path) < 0) {
			perror(path);
			return ERR_DELETE;
		}
	}

	/* remove metadata directory, no matter what */
	snprintf(path, PATH_MAX, "%s/%s", datadir, name);
	log(LOG_DEBUG, "- %s\n", path);
	return rmdir(path);
}


/*
 * Install a pack from the given path. This wraps load/install of a pack
 */
int
install(const char *rootfs, const char *datadir, char *name)
{
	int r = 0;
	char path[PATH_MAX];
	struct pack *p = NULL;

	if (re_match(PACK_FORMAT, name) != 0) {
#ifdef REPOAWARE
		pack_find(name, path);
#else
		fprintf(stderr, "%s: invalid filename\n", name);
		exit(1);
#endif
	} else {
		snprintf(path, PATH_MAX, "%s", name);
	}

	if ((p = pack_load_tarball(path)) == NULL)
		return ERR_PACK_LOAD;

	r += pack_install(rootfs, datadir, p);

	if (r == 0)
		log(LOG_INFO, "installed %s (%s)\n", p->name, p->version);

	pack_free(p);

	return r;
}


/*
 * Update a pack. This should be as easy as delete/install.
 * Deletion is required in case the file structure changes
 */
int
update(const char *rootfs, const char *datadir, char *name)
{
	int r = 0, tmp = overwrite;
	char path[PATH_MAX];
	struct pack *p = NULL;

	if (re_match(PACK_FORMAT, name) != 0) {
#ifdef REPOAWARE
		pack_find(name, path);
#else
		fprintf(stderr, "%s: invalid filename\n", name);
		exit(1);
#endif
	} else {
		snprintf(path, PATH_MAX, "%s", name);
	}

	if ((p = pack_load_tarball(path)) == NULL)
		return ERR_PACK_LOAD;

	if (inspect_installed(datadir, p->name)) {
		fprintf(stderr, "%s: No such pack installed\n", p->name);
		pack_free(p);
		return ERR_DELETE;
	}

	if (pack_delete(rootfs, datadir, p) != 0) {
		pack_free(p);
		return ERR_DELETE;
	}

	overwrite = 1;
	r += pack_install(rootfs, datadir, p);
	overwrite = tmp;

	if (r == 0)
		log(LOG_INFO, "updated %s to %s\n", p->name, p->version);

	pack_free(p);

	return r;
}


/*
 * Delete a currently installed pack. This wraps load/delete functions
 */
int
delete(const char *rootfs, const char *datadir, char *name)
{
	int r = 0;
	struct pack *p = NULL;

	if (inspect_installed(datadir, name)) {
		fprintf(stderr, "%s: No such pack installed\n", name);
		return ERR_DELETE;
	}

	if ((p = pack_load_metadata(datadir, name)) == NULL)
		return ERR_PACK_LOAD;

	r += pack_delete(rootfs, datadir, p);

	if (r == 0)
		log(LOG_INFO, "deleted %s\n", p->name);

	pack_free(p);

	return r;
}


/*
 * Inspect the system, either by looking into a pack, checking installed
 * files, or listing packs actually installed
 */
int
inspect(const char *datadir, const char *packname)
{
	/* name is NULL, list packs installed */
	if (!packname)
		return inspect_system(1, datadir);

	/* otherwise, list files installed by a pack */
	return inspect_files(1, datadir, packname);
}


int
main (int argc, char **argv)
{
	int r = 0;
	char *n = NULL, *argv0;
	uint8_t action = ACTION_DEFAULT;
	char rootfs[PATH_MAX] = "";
	char datadir[PATH_MAX] = "";

	strncpy(rootfs, PACK_ROOT, PATH_MAX);
	strncat(rootfs, "/", PATH_MAX);
	strncpy(datadir, PACK_DATA, PATH_MAX);

	ARGBEGIN{
	case 'a':
		action = ACTION_INSTALL;
		break;
	case 'd':
		action = ACTION_DELETE;
		break;
	case 'f':
		overwrite = 1;
		break;
	case 'i':
		action = ACTION_INSPECT;
		if (argc > 1)
			n = ARGF();
		break;
	case 'u':
		action = ACTION_UPDATE;
		break;
	case 'v':
		verbose++;
		break;
	default:
		action = ACTION_INVALID;
	}ARGEND;

	switch (action) {
	case ACTION_INSTALL:
	case ACTION_UPDATE:
	case ACTION_DELETE:
		while(*argv) {
			if (action == ACTION_INSTALL)
				r += install(rootfs, datadir, *argv);
			if (action == ACTION_UPDATE)
				r += update(rootfs, datadir, *argv);
			if (action == ACTION_DELETE)
				r += delete(rootfs, datadir, *argv);
			argv++;
		}
		break;
	case ACTION_INSPECT:
		if (inspect(datadir, n) != 0)
			return ERR_INSPECT;
		break;
	default:
		usage(argv0);
		return ERR_INVALID_ACTION;
	}

	return r;
}
